@@ -63,10 +63,24 @@ commit_lock_timeout = 30
 class MySQLAdapter(Adapter):
     """MySQL adapter for RelStorage."""
 
+    _scripts = Adapter._scripts.copy()
+    # work around a MySQL performance bug
+    # See: http://mail.zope.org/pipermail/zodb-dev/2008-May/011880.html
+    #      http://bugs.mysql.com/bug.php?id=28257
+    _scripts['prepack_follow_child_refs'] = """
+    UPDATE pack_object SET keep = %(TRUE)s
+    WHERE keep = %(FALSE)s
+        AND zoid IN (
+            SELECT * FROM (
+                SELECT DISTINCT to_zoid
+                FROM object_ref
+                    JOIN temp_pack_visit USING (zoid)
+            ) AS child_zoids
+        )
+    """
+
     def __init__(self, **params):
         self._params = params.copy()
-        self._params['use_unicode'] = True
-        self._params['charset'] = 'utf8'
 
     def create_schema(self, cursor):
         """Create the database tables."""
@@ -75,6 +89,7 @@ class MySQLAdapter(Adapter):
         CREATE TABLE transaction (
             tid         BIGINT NOT NULL PRIMARY KEY,
             packed      BOOLEAN NOT NULL DEFAULT FALSE,
+            empty       BOOLEAN NOT NULL DEFAULT FALSE,
             username    BLOB NOT NULL,
             description BLOB NOT NULL,
             extension   BLOB
@@ -104,6 +119,7 @@ class MySQLAdapter(Adapter):
             CHECK (tid > 0)
         ) ENGINE = InnoDB;
         CREATE INDEX object_state_tid ON object_state (tid);
+        CREATE INDEX object_state_prev_tid ON object_state (prev_tid);
 
         -- Pointers to the current object state
         CREATE TABLE current_object (
@@ -121,11 +137,9 @@ class MySQLAdapter(Adapter):
         CREATE TABLE object_ref (
             zoid        BIGINT NOT NULL,
             tid         BIGINT NOT NULL,
-            to_zoid     BIGINT NOT NULL
+            to_zoid     BIGINT NOT NULL,
+            PRIMARY KEY (tid, zoid, to_zoid)
         ) ENGINE = MyISAM;
-        CREATE INDEX object_ref_from ON object_ref (zoid);
-        CREATE INDEX object_ref_tid ON object_ref (tid);
-        CREATE INDEX object_ref_to ON object_ref (to_zoid);
 
         -- The object_refs_added table tracks whether object_refs has
         -- been populated for all states in a given transaction.
@@ -153,6 +167,19 @@ class MySQLAdapter(Adapter):
             keep_tid    BIGINT
         ) ENGINE = MyISAM;
         CREATE INDEX pack_object_keep_zoid ON pack_object (keep, zoid);
+
+        -- Temporary state during packing: the list of object states to pack.
+        CREATE TABLE pack_state (
+            tid         BIGINT NOT NULL,
+            zoid        BIGINT NOT NULL,
+            PRIMARY KEY (tid, zoid)
+        ) ENGINE = MyISAM;
+
+        -- Temporary state during packing: the list of transactions that
+        -- have at least one object state to pack.
+        CREATE TABLE pack_state_tid (
+            tid         BIGINT NOT NULL PRIMARY KEY
+        ) ENGINE = MyISAM;
         """
         self._run_script(cursor, stmt)
 
@@ -248,8 +275,31 @@ class MySQLAdapter(Adapter):
 
     def get_db_size(self):
         """Returns the approximate size of the database in bytes"""
-        # do later
-        return 0
+        conn, cursor = self.open()
+        try:
+            cursor.execute("SHOW TABLE STATUS")
+            description = [i[0] for i in cursor.description]
+            rows = list(cursor)
+        finally:
+            self.close(conn, cursor)
+        data_column = description.index('Data_length')
+        index_column = description.index('Index_length')
+        return sum([row[data_column] + row[index_column] for row in rows], 0)
+
+    def get_current_tid(self, cursor, oid):
+        """Returns the current integer tid for an object.
+
+        oid is an integer.  Returns None if object does not exist.
+        """
+        cursor.execute("""
+        SELECT tid
+        FROM current_object
+        WHERE zoid = %s
+        """, (oid,))
+        if cursor.rowcount:
+            assert cursor.rowcount == 1
+            return cursor.fetchone()[0]
+        return None
 
     def load_current(self, cursor, oid):
         """Returns the current pickle and integer tid for an object.
@@ -354,11 +404,19 @@ class MySQLAdapter(Adapter):
             self.close(conn, cursor)
             raise
 
+    def _restart_temp_table(self, cursor):
+        """Restart the temporary table for storing objects"""
+        stmt = """
+        DROP TEMPORARY TABLE IF EXISTS temp_store
+        """
+        cursor.execute(stmt)
+        self._make_temp_table(cursor)
+
     def restart_store(self, cursor):
         """Reuse a store connection."""
         try:
             cursor.connection.rollback()
-            cursor.execute("TRUNCATE temp_store")
+            self._restart_temp_table(cursor)
         except (MySQLdb.OperationalError, MySQLdb.InterfaceError), e:
             raise StorageError(e)
 
@@ -398,12 +456,7 @@ class MySQLAdapter(Adapter):
 
     def start_commit(self, cursor):
         """Prepare to commit."""
-        # Hold commit_lock to prevent concurrent commits.
-        cursor.execute("SELECT GET_LOCK('relstorage.commit', %s)",
-            (commit_lock_timeout,))
-        locked = cursor.fetchone()[0]
-        if not locked:
-            raise StorageError("Unable to acquire commit lock")
+        self._hold_commit_lock(cursor)
 
     def get_tid_and_time(self, cursor):
         """Returns the most recent tid and the current database time.
@@ -506,12 +559,12 @@ class MySQLAdapter(Adapter):
     def commit_phase2(self, cursor, txn):
         """Final transaction commit."""
         cursor.connection.commit()
-        cursor.execute("SELECT RELEASE_LOCK('relstorage.commit')")
+        self._release_commit_lock(cursor)
 
     def abort(self, cursor, txn=None):
         """Abort the commit.  If txn is not None, phase 1 is also aborted."""
         cursor.connection.rollback()
-        cursor.execute("SELECT RELEASE_LOCK('relstorage.commit')")
+        self._release_commit_lock(cursor)
 
     def new_oid(self, cursor):
         """Return a new, unused OID."""
@@ -530,7 +583,7 @@ class MySQLAdapter(Adapter):
 
         Raise an exception if packing or undo is already in progress.
         """
-        stmt = "SELECT GET_LOCK('relstorage.pack', 0)"
+        stmt = "SELECT GET_LOCK(CONCAT(DATABASE(), '.pack'), 0)"
         cursor.execute(stmt)
         res = cursor.fetchone()[0]
         if not res:
@@ -539,12 +592,13 @@ class MySQLAdapter(Adapter):
 
     def release_pack_lock(self, cursor):
         """Release the pack lock."""
-        stmt = "SELECT RELEASE_LOCK('relstorage.pack')"
+        stmt = "SELECT RELEASE_LOCK(CONCAT(DATABASE(), '.pack'))"
         cursor.execute(stmt)
 
 
-    def pre_pack(self, pack_tid, get_references, gc):
-        """Decide what to pack.
+    def open_for_pre_pack(self):
+        """Open a connection to be used for the pre-pack phase.
+        Returns (conn, cursor).
 
         This overrides the method by the same name in common.Adapter.
         """
@@ -553,24 +607,30 @@ class MySQLAdapter(Adapter):
             # This phase of packing works best with transactions
             # disabled.  It changes no user-facing data.
             conn.autocommit(True)
-            if gc:
-                self._pre_pack_with_gc(cursor, pack_tid, get_references)
-            else:
-                self._pre_pack_without_gc(cursor, pack_tid)
-        finally:
+            return conn, cursor
+        except:
             self.close(conn, cursor)
+            raise
 
 
     def _hold_commit_lock(self, cursor):
-        """Hold the commit lock for packing.
+        """Hold the commit lock.
 
         This overrides the method by the same name in common.Adapter.
         """
-        cursor.execute("SELECT GET_LOCK('relstorage.commit', %s)",
+        cursor.execute("SELECT GET_LOCK(CONCAT(DATABASE(), '.commit'), %s)",
             (commit_lock_timeout,))
         locked = cursor.fetchone()[0]
         if not locked:
             raise StorageError("Unable to acquire commit lock")
 
 
+    def _release_commit_lock(self, cursor):
+        """Release the commit lock.
+
+        This overrides the method by the same name in common.Adapter.
+        """
+        cursor.execute("SELECT RELEASE_LOCK(CONCAT(DATABASE(), '.commit'))")
+
+
     _poll_query = "SELECT tid FROM transaction ORDER BY tid DESC LIMIT 1"
