@@ -15,8 +15,6 @@
 
 import logging
 import re
-import thread
-import time
 import cx_Oracle
 from ZODB.POSException import StorageError
 
@@ -24,6 +22,27 @@ from common import Adapter
 
 log = logging.getLogger("relstorage.adapters.oracle")
 
+def lob_handler(cursor, name, defaultType, size, precision, scale):
+    """cx_Oracle outputtypehandler that causes Oracle to send BLOBs inline.
+
+    Note that if a BLOB in the result is too large, Oracle generates an
+    error indicating truncation.  The execute_lob_stmt() method works
+    around this.
+    """
+    if defaultType == cx_Oracle.BLOB:
+        # Default size for BLOB is 4, we want the whole blob inline.
+        # Typical chunk size is 8132, we choose a multiple - 32528
+        return cursor.var(cx_Oracle.LONG_BINARY, 32528, cursor.arraysize)
+
+def read_lob(value):
+    """Handle an Oracle LOB by returning its byte stream.
+
+    Returns other objects unchanged.
+    """
+    if isinstance(value, cx_Oracle.LOB):
+        return value.read()
+    return value
+
 
 class OracleAdapter(Adapter):
     """Oracle adapter for RelStorage."""
@@ -32,6 +51,7 @@ class OracleAdapter(Adapter):
         'TRUE':         "'Y'",
         'FALSE':        "'N'",
         'OCTET_LENGTH': 'LENGTH',
+        'TRUNCATE':     'TRUNCATE TABLE',
         'oid':          ':oid',
         'tid':          ':tid',
         'pack_tid':     ':pack_tid',
@@ -61,13 +81,41 @@ class OracleAdapter(Adapter):
         'create_temp_pack_visit': None,
         'create_temp_undo': None,
         'reset_temp_undo': "DELETE FROM temp_undo",
+
+        'transaction_has_data': """
+            SELECT DISTINCT tid
+            FROM object_state
+            WHERE tid = %(tid)s
+            """,
+
+        'prepack_follow_child_refs':
+            Adapter._scripts['prepack_follow_child_refs'],
     }
 
-    def __init__(self, user, password, dsn, twophase=False, arraysize=64):
-        self._params = (user, password, dsn)
-        self._twophase = twophase
-        self._arraysize = arraysize
+    def __init__(self, user, password, dsn, twophase=False, arraysize=64,
+            use_inline_lobs=None):
+        """Create an Oracle adapter.
 
+        The user, password, and dsn parameters are provided to cx_Oracle
+        at connection time.
+
+        If twophase is true, all commits go through an Oracle-level two-phase
+        commit process.  This is disabled by default.  Even when this option
+        is disabled, the ZODB two-phase commit is still in effect.
+
+        arraysize sets the number of rows to buffer in cx_Oracle.  The default
+        is 64.
+
+        use_inline_lobs enables Oracle to send BLOBs inline in response to
+        queries.  It depends on features in cx_Oracle 5.  The default is None,
+        telling the adapter to auto-detect the presence of cx_Oracle 5.
+        """
+        self._params = (user, password, dsn)
+        self._twophase = bool(twophase)
+        self._arraysize = arraysize
+        if use_inline_lobs is None:
+            use_inline_lobs = (cx_Oracle.version >= '5.0')
+        self._use_inline_lobs = bool(use_inline_lobs)
 
     def _run_script_stmt(self, cursor, generic_stmt, generic_params=()):
         """Execute a statement from a script with the given parameters.
@@ -105,6 +153,7 @@ class OracleAdapter(Adapter):
         CREATE TABLE transaction (
             tid         NUMBER(20) NOT NULL PRIMARY KEY,
             packed      CHAR DEFAULT 'N' CHECK (packed IN ('N', 'Y')),
+            empty       CHAR DEFAULT 'N' CHECK (empty IN ('N', 'Y')),
             username    RAW(500),
             description RAW(2000),
             extension   RAW(2000)
@@ -133,6 +182,7 @@ class OracleAdapter(Adapter):
             state       BLOB
         );
         CREATE INDEX object_state_tid ON object_state (tid);
+        CREATE INDEX object_state_prev_tid ON object_state (prev_tid);
 
         -- Pointers to the current object state
         CREATE TABLE current_object (
@@ -140,6 +190,7 @@ class OracleAdapter(Adapter):
             tid         NUMBER(20) NOT NULL,
             FOREIGN KEY (zoid, tid) REFERENCES object_state
         );
+        CREATE INDEX current_object_tid ON current_object (tid);
 
         -- States that will soon be stored
         CREATE GLOBAL TEMPORARY TABLE temp_store (
@@ -161,11 +212,9 @@ class OracleAdapter(Adapter):
         CREATE TABLE object_ref (
             zoid        NUMBER(20) NOT NULL,
             tid         NUMBER(20) NOT NULL,
-            to_zoid     NUMBER(20) NOT NULL
+            to_zoid     NUMBER(20) NOT NULL,
+            PRIMARY KEY (tid, zoid, to_zoid)
         );
-        CREATE INDEX object_ref_from ON object_ref (zoid);
-        CREATE INDEX object_ref_tid ON object_ref (tid);
-        CREATE INDEX object_ref_to ON object_ref (to_zoid);
 
         -- The object_refs_added table tracks whether object_refs has
         -- been populated for all states in a given transaction.
@@ -194,6 +243,19 @@ class OracleAdapter(Adapter):
         );
         CREATE INDEX pack_object_keep_zoid ON pack_object (keep, zoid);
 
+        -- Temporary state during packing: the list of object states to pack.
+        CREATE TABLE pack_state (
+            tid         NUMBER(20) NOT NULL,
+            zoid        NUMBER(20) NOT NULL,
+            PRIMARY KEY (tid, zoid)
+        );
+
+        -- Temporary state during packing: the list of transactions that
+        -- have at least one object state to pack.
+        CREATE TABLE pack_state_tid (
+            tid         NUMBER(20) NOT NULL PRIMARY KEY
+        );
+
         -- Temporary state during packing: a list of objects
         -- whose references need to be examined.
         CREATE GLOBAL TEMPORARY TABLE temp_pack_visit (
@@ -327,39 +389,81 @@ class OracleAdapter(Adapter):
         # May not be possible without access to the dba_* objects
         return 0
 
+    def get_current_tid(self, cursor, oid):
+        """Returns the current integer tid for an object.
+
+        oid is an integer.  Returns None if object does not exist.
+        """
+        cursor.execute("""
+        SELECT tid
+        FROM current_object
+        WHERE zoid = :1
+        """, (oid,))
+        for (tid,) in cursor:
+            return tid
+        return None
+
+    def execute_lob_stmt(self, cursor, stmt, args=(), default=None):
+        """Execute a statement and return one row with all LOBs inline.
+
+        Returns the value of the default parameter if the result was empty.
+        """
+        if self._use_inline_lobs:
+            try:
+                cursor.outputtypehandler = lob_handler
+                try:
+                    cursor.execute(stmt, args)
+                    for row in cursor:
+                        return row
+                finally:
+                    del cursor.outputtypehandler
+            except cx_Oracle.DatabaseError, e:
+                # ORA-01406: fetched column value was truncated
+                error, = e
+                if ((isinstance(error, str) and not error.endswith(' 1406'))
+                        or error.code != 1406):
+                    raise
+                # Execute the query, but alter it slightly without
+                # changing its meaning, so that the query cache
+                # will see it as a statement that has to be compiled
+                # with different output type parameters.
+                cursor.execute(stmt + ' ', args)
+                for row in cursor:
+                    return tuple(map(read_lob, row))
+        else:
+            cursor.execute(stmt, args)
+            for row in cursor:
+                return tuple(map(read_lob, row))
+        return default
+
     def load_current(self, cursor, oid):
         """Returns the current pickle and integer tid for an object.
 
         oid is an integer.  Returns (None, None) if object does not exist.
         """
-        cursor.execute("""
+        stmt = """
         SELECT state, tid
         FROM current_object
             JOIN object_state USING(zoid, tid)
         WHERE zoid = :1
-        """, (oid,))
-        for state, tid in cursor:
-            if state is not None:
-                state = state.read()
-            # else this object's creation has been undone
-            return state, tid
-        return None, None
+        """
+        return self.execute_lob_stmt(
+            cursor, stmt, (oid,), default=(None, None))
 
     def load_revision(self, cursor, oid, tid):
         """Returns the pickle for an object on a particular transaction.
 
         Returns None if no such state exists.
         """
-        cursor.execute("""
+        stmt = """
         SELECT state
         FROM object_state
         WHERE zoid = :1
             AND tid = :2
-        """, (oid, tid))
-        for (state,) in cursor:
-            if state is not None:
-                return state.read()
-        return None
+        """
+        (state,) = self.execute_lob_stmt(
+            cursor, stmt, (oid, tid), default=(None,))
+        return state
 
     def exists(self, cursor, oid):
         """Returns a true value if the given object exists."""
@@ -382,13 +486,8 @@ class OracleAdapter(Adapter):
                     AND tid < :tid
             )
         """
-        cursor.execute(stmt, {'oid': oid, 'tid': tid})
-        for state, tid in cursor:
-            if state is not None:
-                state = state.read()
-            # else this object's creation has been undone
-            return state, tid
-        return None, None
+        return self.execute_lob_stmt(cursor, stmt, {'oid': oid, 'tid': tid},
+             default=(None, None))
 
     def get_object_tid_after(self, cursor, oid, tid):
         """Returns the tid of the next change after an object revision.
@@ -539,13 +638,10 @@ class OracleAdapter(Adapter):
             JOIN current_object ON (temp_store.zoid = current_object.zoid)
         WHERE temp_store.prev_tid != current_object.tid
         """
-        cursor.execute(stmt)
-        for oid, prev_tid, attempted_prev_tid, data in cursor:
-            return oid, prev_tid, attempted_prev_tid, data.read()
-        return None
+        return self.execute_lob_stmt(cursor, stmt)
 
     def move_from_temp(self, cursor, tid):
-        """Moved the temporarily stored objects to permanent storage.
+        """Move the temporarily stored objects to permanent storage.
 
         Returns the list of oids stored.
         """
