@@ -16,14 +16,13 @@
 from base64 import decodestring, encodestring
 import logging
 import psycopg2, psycopg2.extensions
+import re
 from ZODB.POSException import StorageError
 
 from common import Adapter
 
 log = logging.getLogger("relstorage.adapters.postgresql")
 
-psycopg2.extensions.register_type(psycopg2.extensions.UNICODE)
-
 
 class PostgreSQLAdapter(Adapter):
     """PostgreSQL adapter for RelStorage."""
@@ -41,6 +40,7 @@ class PostgreSQLAdapter(Adapter):
         CREATE TABLE transaction (
             tid         BIGINT NOT NULL PRIMARY KEY,
             packed      BOOLEAN NOT NULL DEFAULT FALSE,
+            empty       BOOLEAN NOT NULL DEFAULT FALSE,
             username    BYTEA NOT NULL,
             description BYTEA NOT NULL,
             extension   BYTEA
@@ -66,6 +66,7 @@ class PostgreSQLAdapter(Adapter):
             state       BYTEA
         );
         CREATE INDEX object_state_tid ON object_state (tid);
+        CREATE INDEX object_state_prev_tid ON object_state (prev_tid);
 
         -- Pointers to the current object state
         CREATE TABLE current_object (
@@ -73,9 +74,7 @@ class PostgreSQLAdapter(Adapter):
             tid         BIGINT NOT NULL,
             FOREIGN KEY (zoid, tid) REFERENCES object_state
         );
-
-        -- During packing, an exclusive lock is held on pack_lock.
-        CREATE TABLE pack_lock ();
+        CREATE INDEX current_object_tid ON current_object (tid);
 
         -- A list of referenced OIDs from each object_state.
         -- This table is populated as needed during packing.
@@ -86,11 +85,9 @@ class PostgreSQLAdapter(Adapter):
         CREATE TABLE object_ref (
             zoid        BIGINT NOT NULL,
             tid         BIGINT NOT NULL,
-            to_zoid     BIGINT NOT NULL
+            to_zoid     BIGINT NOT NULL,
+            PRIMARY KEY (tid, zoid, to_zoid)
         );
-        CREATE INDEX object_ref_from ON object_ref (zoid);
-        CREATE INDEX object_ref_tid ON object_ref (tid);
-        CREATE INDEX object_ref_to ON object_ref (to_zoid);
 
         -- The object_refs_added table tracks whether object_refs has
         -- been populated for all states in a given transaction.
@@ -117,10 +114,29 @@ class PostgreSQLAdapter(Adapter):
             keep        BOOLEAN NOT NULL,
             keep_tid    BIGINT
         );
-        CREATE INDEX pack_object_keep_zoid ON pack_object (keep, zoid);
+        CREATE INDEX pack_object_keep_false ON pack_object (zoid)
+            WHERE keep = false;
+        CREATE INDEX pack_object_keep_true ON pack_object (zoid, keep_tid)
+            WHERE keep = true;
+
+        -- Temporary state during packing: the list of object states to pack.
+        CREATE TABLE pack_state (
+            tid         BIGINT NOT NULL,
+            zoid        BIGINT NOT NULL,
+            PRIMARY KEY (tid, zoid)
+        );
+
+        -- Temporary state during packing: the list of transactions that
+        -- have at least one object state to pack.
+        CREATE TABLE pack_state_tid (
+            tid         BIGINT NOT NULL PRIMARY KEY
+        );
         """
         cursor.execute(stmt)
 
+        if not self._pg_has_advisory_locks(cursor):
+            cursor.execute("CREATE TABLE pack_lock ()")
+
 
     def prepare_schema(self):
         """Create the database schema if it does not already exist."""
@@ -173,7 +189,6 @@ class PostgreSQLAdapter(Adapter):
         """Open a database connection and return (conn, cursor)."""
         try:
             conn = psycopg2.connect(self._dsn)
-            conn.set_client_encoding('UNICODE')
             conn.set_isolation_level(isolation)
             cursor = conn.cursor()
             cursor.arraysize = 64
@@ -193,6 +208,20 @@ class PostgreSQLAdapter(Adapter):
                         psycopg2.OperationalError):
                     pass
 
+    def _pg_version(self, cursor):
+        """Return the (major, minor) version of PostgreSQL"""
+        cursor.execute("SELECT version()")
+        v = cursor.fetchone()[0]
+        m = re.search(r"([0-9]+)[.]([0-9]+)", v)
+        if m is None:
+            raise AssertionError("Unable to detect PostgreSQL version: " + v)
+        else:
+            return int(m.group(1)), int(m.group(2))
+
+    def _pg_has_advisory_locks(self, cursor):
+        """Return true if this version of PostgreSQL supports advisory locks"""
+        return self._pg_version(cursor) >= (8, 2)
+
     def open_for_load(self):
         """Open and initialize a connection for loading objects.
 
@@ -231,6 +260,21 @@ class PostgreSQLAdapter(Adapter):
         finally:
             self.close(conn, cursor)
 
+    def get_current_tid(self, cursor, oid):
+        """Returns the current integer tid for an object.
+
+        oid is an integer.  Returns None if object does not exist.
+        """
+        cursor.execute("""
+        SELECT tid
+        FROM current_object
+        WHERE zoid = %s
+        """, (oid,))
+        if cursor.rowcount:
+            assert cursor.rowcount == 1
+            return cursor.fetchone()[0]
+        return None
+
     def load_current(self, cursor, oid):
         """Returns the current pickle and integer tid for an object.
 
@@ -556,23 +600,27 @@ class PostgreSQLAdapter(Adapter):
         cursor.execute(stmt)
         return cursor.fetchone()[0]
 
-
     def hold_pack_lock(self, cursor):
         """Try to acquire the pack lock.
 
         Raise an exception if packing or undo is already in progress.
         """
-        stmt = """
-        LOCK pack_lock IN EXCLUSIVE MODE NOWAIT
-        """
-        try:
-            cursor.execute(stmt)
-        except psycopg2.DatabaseError:
-            raise StorageError('A pack or undo operation is in progress')
+        if self._pg_has_advisory_locks(cursor):
+            cursor.execute("SELECT pg_try_advisory_lock(1)")
+            locked = cursor.fetchone()[0]
+            if not locked:
+                raise StorageError('A pack or undo operation is in progress')
+        else:
+            # b/w compat
+            try:
+                cursor.execute("LOCK pack_lock IN EXCLUSIVE MODE NOWAIT")
+            except psycopg2.DatabaseError:
+                raise StorageError('A pack or undo operation is in progress')
 
     def release_pack_lock(self, cursor):
         """Release the pack lock."""
-        # No action needed
-        pass
+        if self._pg_has_advisory_locks(cursor):
+            cursor.execute("SELECT pg_advisory_unlock(1)")
+        # else no action needed since the lock will be released at txn commit
 
     _poll_query = "EXECUTE get_latest_tid"
