@@ -23,17 +23,26 @@ import md5
 import os
 import time
 import weakref
-from ZODB.utils import p64, u64, z64
-from ZODB.BaseStorage import BaseStorage
+from ZODB.utils import p64, u64
+from ZODB.BaseStorage import BaseStorage, TransactionRecord, DataRecord
 from ZODB import ConflictResolution, POSException
 from persistent.TimeStamp import TimeStamp
 
+try:
+    from ZODB.interfaces import StorageStopIteration
+except ImportError:
+    class StorageStopIteration(IndexError, StopIteration):
+        """A combination of StopIteration and IndexError to provide a
+        backwards-compatible exception.
+        """
+
+
 log = logging.getLogger("relstorage")
 
 # Set the RELSTORAGE_ABORT_EARLY environment variable when debugging
 # a failure revealed by the ZODB test suite.  The test suite often fails
 # to call tpc_abort in the event of an error, leading to deadlocks.
-# The variable causes RelStorage to abort failed transactions
+# This variable causes RelStorage to abort failed transactions
 # early rather than wait for an explicit abort.
 abort_early = os.environ.get('RELSTORAGE_ABORT_EARLY')
 
@@ -43,15 +52,17 @@ class RelStorage(BaseStorage,
     """Storage to a relational database, based on invalidation polling"""
 
     def __init__(self, adapter, name=None, create=True,
-            read_only=False, poll_interval=0, pack_gc=True):
+            read_only=False, options=None):
         if name is None:
             name = 'RelStorage on %s' % adapter.__class__.__name__
 
         self._adapter = adapter
         self._name = name
         self._is_read_only = read_only
-        self._poll_interval = poll_interval
-        self._pack_gc = pack_gc
+        if options is None:
+            options = Options()
+        self._options = options
+        self._cache_client = None
 
         if create:
             self._adapter.prepare_schema()
@@ -91,6 +102,17 @@ class RelStorage(BaseStorage,
         # _max_new_oid is the highest OID provided by new_oid()
         self._max_new_oid = 0
 
+        # set _cache_client
+        if options.cache_servers:
+            module_name = options.cache_module_name
+            module = __import__(module_name, {}, {}, ['Client'])
+            servers = options.cache_servers
+            if isinstance(servers, basestring):
+                servers = servers.split()
+            self._cache_client = module.Client(servers)
+        else:
+            self._cache_client = None
+
 
     def _open_load_connection(self):
         """Open the load connection to the database.  Return nothing."""
@@ -173,6 +195,9 @@ class RelStorage(BaseStorage,
         """
         self._adapter.zap_all()
         self._rollback_load_connection()
+        cache = self._cache_client
+        if cache is not None:
+            cache.flush_all()
 
     def close(self):
         """Close the connections to the database."""
@@ -210,24 +235,89 @@ class RelStorage(BaseStorage,
         """Return database size in bytes"""
         return self._adapter.get_db_size()
 
+    def _get_oid_cache_key(self, oid_int):
+        """Return the cache key for finding the current tid.
+
+        This is overridden by BoundRelStorage.
+        """
+        return None
+
+    def _log_keyerror(self, oid_int, reason):
+        """Log just before raising KeyError in load().
+
+        KeyErrors in load() are generally not supposed to happen,
+        so this is a good place to gather information.
+        """
+        cursor = self._load_cursor
+        adapter = self._adapter
+        msg = ["Storage KeyError on oid %d: %s" % (oid_int, reason)]
+        rows = adapter.iter_transactions(cursor)
+        row = None
+        for row in rows:
+            # just get the first row
+            break
+        if not row:
+            msg.append("No transactions exist")
+        else:
+            msg.append("Current transaction is %d" % row[0])
+
+        rows = adapter.iter_object_history(cursor, oid_int)
+        tids = []
+        for row in rows:
+            tids.append(row[0])
+            if len(tids) >= 10:
+                break
+        msg.append("Recent object tids: %s" % repr(tids))
+        log.warning('; '.join(msg))
+
     def load(self, oid, version):
+        oid_int = u64(oid)
+        cache = self._cache_client
+
         self._lock_acquire()
         try:
             if not self._load_transaction_open:
                 self._restart_load()
             cursor = self._load_cursor
-            state, tid_int = self._adapter.load_current(cursor, u64(oid))
+            if cache is None:
+                state, tid_int = self._adapter.load_current(cursor, oid_int)
+            else:
+                # get tid_int from the cache or the database
+                cachekey = self._get_oid_cache_key(oid_int)
+                if cachekey:
+                    tid_int = cache.get(cachekey)
+                if not cachekey or not tid_int:
+                    tid_int = self._adapter.get_current_tid(
+                        cursor, oid_int)
+                    if cachekey and tid_int is not None:
+                        cache.set(cachekey, tid_int)
+                if tid_int is None:
+                    self._log_keyerror(oid_int, "no tid found(1)")
+                    raise KeyError(oid)
+
+                # get state from the cache or the database
+                cachekey = 'state:%d:%d' % (oid_int, tid_int)
+                state = cache.get(cachekey)
+                if not state:
+                    state = self._adapter.load_revision(
+                        cursor, oid_int, tid_int)
+                    if state:
+                        state = str(state)
+                        cache.set(cachekey, state)
         finally:
             self._lock_release()
+
         if tid_int is not None:
             if state:
                 state = str(state)
             if not state:
                 # This can happen if something attempts to load
                 # an object whose creation has been undone.
+                self._log_keyerror(oid_int, "creation has been undone")
                 raise KeyError(oid)
             return state, p64(tid_int)
         else:
+            self._log_keyerror(oid_int, "no tid found(2)")
             raise KeyError(oid)
 
     def loadEx(self, oid, version):
@@ -237,6 +327,15 @@ class RelStorage(BaseStorage,
 
     def loadSerial(self, oid, serial):
         """Load a specific revision of an object"""
+        oid_int = u64(oid)
+        tid_int = u64(serial)
+        cache = self._cache_client
+        if cache is not None:
+            cachekey = 'state:%d:%d' % (oid_int, tid_int)
+            state = cache.get(cachekey)
+            if state:
+                return state
+
         self._lock_acquire()
         try:
             if self._store_cursor is not None:
@@ -247,17 +346,20 @@ class RelStorage(BaseStorage,
                 if not self._load_transaction_open:
                     self._restart_load()
                 cursor = self._load_cursor
-            state = self._adapter.load_revision(cursor, u64(oid), u64(serial))
-            if state is not None:
-                state = str(state)
-                if not state:
-                    raise POSKeyError(oid)
-                return state
-            else:
-                raise KeyError(oid)
+            state = self._adapter.load_revision(cursor, oid_int, tid_int)
         finally:
             self._lock_release()
 
+        if state is not None:
+            state = str(state)
+            if not state:
+                raise POSKeyError(oid)
+            if cache is not None:
+                cache.set(cachekey, state)
+            return state
+        else:
+            raise KeyError(oid)
+
     def loadBefore(self, oid, tid):
         """Return the most recent revision of oid before tid committed."""
         oid_int = u64(oid)
@@ -608,8 +710,6 @@ class RelStorage(BaseStorage,
                      'time': TimeStamp(tid).timeTime(),
                      'user_name': user or '',
                      'description': desc or ''}
-                if ext:
-                    ext = str(ext)
                 if ext:
                     d.update(cPickle.loads(ext))
                 if filter is None or filter(d):
@@ -704,22 +804,6 @@ class RelStorage(BaseStorage,
             self._lock_release()
 
 
-    def set_pack_gc(self, pack_gc):
-        """Configures whether garbage collection during packing is enabled.
-
-        Garbage collection is enabled by default.  If GC is disabled,
-        packing keeps at least one revision of every object.
-        With GC disabled, the pack code does not need to follow object
-        references, making packing conceivably much faster.
-        However, some of that benefit may be lost due to an ever
-        increasing number of unused objects.
-
-        Disabling garbage collection is also a hack that ensures
-        inter-database references never break.
-        """
-        self._pack_gc = pack_gc
-
-
     def pack(self, t, referencesf):
         if self._is_read_only:
             raise POSException.ReadOnlyError()
@@ -753,10 +837,10 @@ class RelStorage(BaseStorage,
                 # In pre_pack, the adapter fills tables with
                 # information about what to pack.  The adapter
                 # should not actually pack anything yet.
-                adapter.pre_pack(tid_int, get_references, self._pack_gc)
+                adapter.pre_pack(tid_int, get_references, self._options)
 
                 # Now pack.
-                adapter.pack(tid_int)
+                adapter.pack(tid_int, self._options)
                 self._after_pack()
             finally:
                 adapter.release_pack_lock(lock_cursor)
@@ -787,14 +871,20 @@ class BoundRelStorage(RelStorage):
         # self._zodb_conn = zodb_conn
         RelStorage.__init__(self, adapter=parent._adapter, name=parent._name,
             create=False, read_only=parent._is_read_only,
-            poll_interval=parent._poll_interval, pack_gc=parent._pack_gc)
+            options=parent._options)
         # _prev_polled_tid contains the tid at the previous poll
         self._prev_polled_tid = None
         self._poll_at = 0
 
+    def _get_oid_cache_key(self, oid_int):
+        my_tid = self._prev_polled_tid
+        if my_tid is None:
+            return None
+        return 'tid:%d:%d' % (oid_int, my_tid)
+
     def connection_closing(self):
         """Release resources."""
-        if not self._poll_interval:
+        if not self._options.poll_interval:
             self._rollback_load_connection()
         # else keep the load transaction open so that it's possible
         # to ignore the next poll.
@@ -820,7 +910,7 @@ class BoundRelStorage(RelStorage):
             if self._closed:
                 return {}
 
-            if self._poll_interval:
+            if self._options.poll_interval:
                 now = time.time()
                 if self._load_transaction_open and now < self._poll_at:
                     # It's not yet time to poll again.  The previous load
@@ -828,7 +918,7 @@ class BoundRelStorage(RelStorage):
                     # ignore this poll.
                     return {}
                 # else poll now after resetting the timeout
-                self._poll_at = now + self._poll_interval
+                self._poll_at = now + self._options.poll_interval
 
             self._restart_load()
             conn = self._load_conn
@@ -891,6 +981,9 @@ class TransactionIterator(object):
     def iterator(self):
         return self
 
+    def __iter__(self):
+        return self
+
     def __len__(self):
         return len(self._transactions)
 
@@ -901,29 +994,54 @@ class TransactionIterator(object):
     def next(self):
         if self._closed:
             raise IOError("TransactionIterator already closed")
+        if self._index >= len(self._transactions):
+            raise StorageStopIteration()
         params = self._transactions[self._index]
-        res = RecordIterator(self, *params)
+        res = RelStorageTransactionRecord(self, *params)
         self._index += 1
         return res
 
 
-class RecordIterator(object):
-    """Iterate over the objects in a transaction."""
+class RelStorageTransactionRecord(TransactionRecord):
+
     def __init__(self, trans_iter, tid_int, user, desc, ext, packed):
+        self._trans_iter = trans_iter
+        self._tid_int = tid_int
         self.tid = p64(tid_int)
         self.status = packed and 'p' or ' '
         self.user = user or ''
         self.description = desc or ''
         if ext:
-            self._extension = cPickle.loads(ext)
+            self.extension = cPickle.loads(ext)
         else:
-            self._extension = {}
+            self.extension = {}
 
-        cursor = trans_iter._cursor
-        adapter = trans_iter._adapter
+    # maintain compatibility with the old (ZODB 3.8 and below) name of
+    # the extension attribute.
+    def _ext_set(self, value):
+        self.extension = value
+    def _ext_get(self):
+        return self.extension
+    _extension = property(fset=_ext_set, fget=_ext_get)
+
+    def __iter__(self):
+        return RecordIterator(self)
+
+
+class RecordIterator(object):
+    """Iterate over the objects in a transaction."""
+    def __init__(self, record):
+        # record is a RelStorageTransactionRecord.
+        cursor = record._trans_iter._cursor
+        adapter = record._trans_iter._adapter
+        tid_int = record._tid_int
+        self.tid = record.tid
         self._records = list(adapter.iter_objects(cursor, tid_int))
         self._index = 0
 
+    def __iter__(self):
+        return self
+
     def __len__(self):
         return len(self._records)
 
@@ -932,13 +1050,15 @@ class RecordIterator(object):
         return self.next()
 
     def next(self):
+        if self._index >= len(self._records):
+            raise StorageStopIteration()
         params = self._records[self._index]
         res = Record(self.tid, *params)
         self._index += 1
         return res
 
 
-class Record(object):
+class Record(DataRecord):
     """An object state in a transaction"""
     version = ''
     data_txn = None
@@ -951,3 +1071,14 @@ class Record(object):
         else:
             self.data = None
 
+
+class Options:
+    """Options for tuning RelStorage."""
+    def __init__(self):
+        self.poll_interval = 0
+        self.pack_gc = True
+        self.pack_batch_timeout = 5.0
+        self.pack_duty_cycle = 0.5
+        self.pack_max_delay = 20.0
+        self.cache_servers = ()  # ['127.0.0.1:11211']
+        self.cache_module_name = 'memcache'
